import argparse
import random
import string

import numpy as np
import torch
import torch.utils.data
import torchvision
from captcha.image import ImageCaptcha
from matplotlib import pyplot as plt
from tqdm import tqdm

plt.rcParams["figure.figsize"] = (15, 5)
plt.style.use('dark_background')

parser = argparse.ArgumentParser()

parser.add_argument('-device', type=str, default='cpu')
parser.add_argument('-batch_size', type=int, default=16)
parser.add_argument('-learning_rate', type=float, default=1e-3)

parser.add_argument('-input_captcha_width', type=int, default=128)
parser.add_argument('-input_captcha_height', type=int, default=64)
parser.add_argument('-input_captcha_length', type=int, default=4)

parser.add_argument('-num_epochs', type=int, default=10)
parser.add_argument('-input_captcha_samples_per_epoch', type=int, default=1000)

args = parser.parse_args()

class Dataset(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__()
        self.characters = string.digits + string.ascii_uppercase
        self.generator = ImageCaptcha(
            args.input_captcha_width,
            args.input_captcha_height
        )

    def __len__(self):
        return args.input_captcha_samples_per_epoch

    def __getitem__(self, idx):

        y_str = ''.join([random.choice(self.characters) for _ in range(args.input_captcha_length)])
        x = np.array(self.generator.generate_image(y_str)) / 255.0
        y_idxes = [self.characters.find(x) for x in y_str]
        x = x.swapaxes(0, 2).astype(np.float32)

        y = np.zeros((args.input_captcha_length, len(self.characters)), dtype=np.float32)
        for i, y_idx in enumerate(y_idxes):
            y[i, y_idx] = 1.0

        return x, np.array(y_idxes, dtype=np.long), y


dataset_full = Dataset()

dataloader_train = torch.utils.data.DataLoader(
    dataset=dataset_full,
    batch_size=args.batch_size,
    shuffle=True,
    drop_last=(len(dataset_full) % args.batch_size < args.batch_size * 0.7)
)

class Model(torch.nn.Module):
    def __init__(self):
        super().__init__()
        # TODO implement ResNet, DenseNet or ConvNet model
        self.encoder = torch.nn.Sequential(
            torch.nn.Conv2d(in_channels=3, out_channels=6, kernel_size=3, padding=1, stride=1)
        )

    def forward(self, x):
        # TODO change following lines
        y_prim = torch.zeros((x.size(0), args.input_captcha_length, len(dataset_full.characters)))
        return y_prim


model = Model()
model = model.to(args.device)
optimizer = torch.optim.Adam(model.parameters(), lr=args.learning_rate)

for epoch in range(1, args.num_epochs + 1):
    losses = []
    accs = []
    for x, y_idxes, y in tqdm(dataloader_train, desc=f'epoch: {epoch}'):
        x = x.to(args.device)
        y = y.to(args.device)

        y_prim = model.forward(x)
        y_prim_idxes = y_prim.argmax(-1)

        # TODO implement loss
        loss = torch.FloatTensor([0])
        # loss.backward()
        # optimizer.step()
        # optimizer.zero_grad()

        # TODO implement acc
        acc = torch.FloatTensor([0])

        accs.append(acc.item())
        losses.append(loss.item())

    loss_epoch = np.mean(losses)
    acc_epoch = np.mean(accs)

    print(f'epoch {epoch}: loss {loss_epoch} acc: {acc_epoch}')

    plt.clf()
    plt.subplot(121)  # row col idx
    plts = []
    c = 0
    for key, value in [
        ('loss', losses),
        ('acc', accs)
    ]:
        plts += plt.plot(value, f'C{c}', label=key)
        ax = plt.twinx()
        c += 1

    plt.legend(plts, [it.get_label() for it in plts])

    for i, j in enumerate([4, 5, 6, 10, 11, 12, 16, 17, 18]):
        plt.subplot(3, 6, j)

        plt.imshow(x[i].cpu().numpy().swapaxes(0, 2))
        y_str = ''.join([dataset_full.characters[it] for it in y_idxes[i].cpu().numpy()])
        y_prim_str = ''.join([dataset_full.characters[it] for it in y_prim_idxes[i].cpu().numpy()])
        plt.title(y_str + '\n' + y_prim_str)

    plt.tight_layout(pad=0.5)
    plt.show()
